package com.ghomerr.linkedchest;

import java.util.HashMap;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;

import org.bukkit.ChatColor;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.block.Chest;
import org.bukkit.entity.Player;
import org.bukkit.inventory.Inventory;
import org.bukkit.plugin.java.JavaPlugin;

import com.ghomerr.linkedchest.constants.Constants;
import com.ghomerr.linkedchest.enums.Commands;
import com.ghomerr.linkedchest.enums.Configurations;
import com.ghomerr.linkedchest.enums.Messages;
import com.ghomerr.linkedchest.listeners.BlockEventListener;
import com.ghomerr.linkedchest.listeners.CommandsListener;
import com.ghomerr.linkedchest.listeners.InventoryInteractionListener;
import com.ghomerr.linkedchest.objects.VirtualInventory;
import com.ghomerr.linkedchest.utils.ConfigurationUtils;
import com.ghomerr.linkedchest.utils.FileUtils;
import com.ghomerr.linkedchest.utils.MessagesUtils;
import com.ghomerr.linkedchest.utils.PermissionsUtils;
import com.ghomerr.linkedchest.utils.StringUtils;
import com.ghomerr.linkedchest.utils.WorldUtils;

public class LinkedChest extends JavaPlugin
{
	private static final Logger _LOGGER = Logger.getLogger(Constants.MINECRAFT);

	public InventoryInteractionListener invListener = null;
	public BlockEventListener blockListener = null;

	private Properties _pluginData = null;
	private Object _dataMonitor = new Object();

	private HashMap<String, VirtualInventory> _virtualInventoriesMap = new HashMap<String, VirtualInventory>();
	private HashMap<String, String> _masterChestsShortLocationsMap = new HashMap<String, String>();
	private HashMap<String, String> _masterDoubleChestsShortLocationsMap = new HashMap<String, String>();
	private Object _mapsMonitor = new Object();

	/* JAVAPLUGIN METHODS */

	@Override
	public void onEnable()
	{
		super.onEnable();

		// Load LinkedChest Config
		final boolean isConfigLoaded = ConfigurationUtils.loadConfiguration();

		// Load LinkedChest Data
		_pluginData = FileUtils.loadData(Constants.DATA_FILE_PATH);
		final boolean isDataLoaded = loadLinkedChestData();

		if (isConfigLoaded && _pluginData != null && isDataLoaded)
		{
			// Load Messages
			MessagesUtils.loadMessages(getLanguage());
			
			// Load Permissions
			PermissionsUtils.loadPermissions(this);

			// Register Commands
			final CommandsListener cmdExec = new CommandsListener(this);
			Commands.registerMainCommand(this, cmdExec);

			// Register Events
			invListener = new InventoryInteractionListener(this);
			blockListener = new BlockEventListener(this);
		}
		else
		{
			_LOGGER.severe(Constants.TAG + "Failed to load Plugin data. Plugin disabled.");
		}
	}

	@Override
	public void onDisable()
	{
		super.onDisable();
		FileUtils.saveFile(_pluginData, Constants.DATA_FILE_PATH);
	}

	/* PUBLIC METHODS */

	public Inventory getVirtualInventory(final String name)
	{
		VirtualInventory virtualInv = null;

		if (StringUtils.isValidChestName(name))
		{
			final String chestName = name.toLowerCase();
			synchronized (_mapsMonitor)
			{
				virtualInv = _virtualInventoriesMap.get(chestName);
			}

			if (virtualInv != null)
			{
				if (virtualInv.inventory == null)
				{
					final Location masterChestLocation = virtualInv.location;
					final World world = masterChestLocation.getWorld();
					final Block block = world.getBlockAt(masterChestLocation);

					if (WorldUtils.isChest(block))
					{
						final Chest chest = (Chest) block.getState();
						synchronized (_mapsMonitor)
						{
							virtualInv.inventory = chest.getInventory();
						}
					}
					else
					{
						_LOGGER.severe(Constants.TAG + "Master Chest " + chestName
								+ " does not exist anymore. Removing it from data.");
						removeVirtualInventory(chestName);
					}
				}
				return virtualInv.inventory;
			}
		}

		return null;
	}

	public String getMasterChestNameFromBlock(final Block block)
	{
		Location loc = block.getLocation();
		
		final String shortLoc = StringUtils.printShortLocation(loc);

		String chestName = null;

		synchronized (_mapsMonitor)
		{
			chestName = _masterChestsShortLocationsMap.get(shortLoc);
		}

		if (chestName == null)
		{
			synchronized (_mapsMonitor)
			{
				chestName = _masterDoubleChestsShortLocationsMap.get(shortLoc);
			}
			
			if (chestName == null)
			{
				String chestShortLoc = null;
				Location nearbyChestLoc = WorldUtils.getChestNearby(block, Location.class);

				if (nearbyChestLoc != null)
				{
					chestShortLoc = StringUtils.printShortLocation(nearbyChestLoc);
					
					synchronized (_mapsMonitor)
					{
						chestName = _masterChestsShortLocationsMap.get(chestShortLoc);
					}
					
					addDoubleChestLocation(shortLoc, chestName);
				}
			}
		}

		return chestName;
	}
	
	public void addDoubleChestLocation(final Location loc, final String chestName)
	{
		addDoubleChestLocation(StringUtils.printShortLocation(loc), chestName);
	}
	
	public void addDoubleChestLocation(final String shortLocation, final String chestName)
	{
		if (chestName != null)
		{
			synchronized (_mapsMonitor)
			{
				_masterDoubleChestsShortLocationsMap.put(shortLocation, chestName.toLowerCase());
			}
		}
	}

	public void setVirtualInventory(final String name, final VirtualInventory virtualInv)
	{
		synchronized (_mapsMonitor)
		{
			_virtualInventoriesMap.put(name, virtualInv);
			if (virtualInv != null)
			{
				_masterChestsShortLocationsMap.put(StringUtils.printShortLocation(virtualInv.location), name);
			}
		}
	}

	public boolean removeVirtualInventory(final String name)
	{
		VirtualInventory removedInv = null;
		final String chestName = name.toLowerCase();
		
		synchronized (_mapsMonitor)
		{
			removedInv = _virtualInventoriesMap.remove(chestName);
			if (removedInv != null)
			{
				final String shortLoc = StringUtils.printShortLocation(removedInv.location);
				_masterChestsShortLocationsMap.remove(shortLoc);
				
				final Block masterChest = removedInv.location.getWorld().getBlockAt(removedInv.location);
				Block doubleChest = WorldUtils.getChestNearby(masterChest, Block.class);
				if (doubleChest != null)
				{
					final String doubleChestShortLoc = StringUtils.printShortLocation(doubleChest.getLocation());
					_masterDoubleChestsShortLocationsMap.remove(doubleChestShortLoc);
				}
			}
		}

		synchronized (_dataMonitor)
		{
			_pluginData.remove(chestName);
			FileUtils.saveFile(_pluginData, Constants.DATA_FILE_PATH);
		}

		return (removedInv != null);
	}

	public boolean addVirtualInventory(final String chestName, final Block block, final boolean isAdmin)
	{
		if (StringUtils.isValidChestName(chestName))
		{
			final Chest chest = (Chest) block.getState();
			final Inventory chestInv = chest.getInventory();

			final VirtualInventory virtualInv = new VirtualInventory(chestInv, chest.getLocation(), isAdmin);
			final String lowerChestName = chestName.toLowerCase();
			setVirtualInventory(lowerChestName, virtualInv);

			synchronized (_dataMonitor)
			{
				_pluginData.put(lowerChestName, virtualInv.getDataString());
				FileUtils.saveFile(_pluginData, Constants.DATA_FILE_PATH);
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	public boolean openVirtualInventory(final Player player, final String chestName)
	{
		final Inventory virtualInv = getVirtualInventory(chestName);
		if (virtualInv != null)
		{
			player.openInventory(virtualInv);
			return true;
		}
		else
		{
			return false;
		}
	}

	public String getLinkedChestsList()
	{
		final StringBuilder strBld = new StringBuilder(MessagesUtils.getWithColor(Messages.AVAILABLE_CHESTS,
				ChatColor.YELLOW));
		boolean isFirst = true;
		Set<String> chestsNamesList = null;

		synchronized (_mapsMonitor)
		{
			chestsNamesList = _virtualInventoriesMap.keySet();
		}

		for (final String chestName : chestsNamesList)
		{
			if (!isFirst)
			{
				strBld.append(ChatColor.YELLOW).append(Constants.DELIMITER).append(Constants.EMPTY_STRING);
			}
			else
			{
				isFirst = false;
			}
			strBld.append(ChatColor.AQUA).append(chestName);
		}

		return strBld.toString();
	}

	public boolean isAdminChest(final String name)
	{
		final String chestName = name.toLowerCase();
		
		VirtualInventory vInv = null;
		synchronized(_mapsMonitor)
		{	
			vInv = _virtualInventoriesMap.get(chestName);
		}
		
		if (vInv != null)
		{
			return vInv.isAdmin;
		}
		else
		{
			return false;
		}
	}
	
	public boolean toggleAdminChest(final String name) throws IllegalArgumentException
	{
		final String chestName = name.toLowerCase();
		
		VirtualInventory vInv = null;
		synchronized(_mapsMonitor)
		{	
			vInv = _virtualInventoriesMap.get(chestName);
			if (vInv != null)
			{
				vInv.isAdmin = !vInv.isAdmin;
				return vInv.isAdmin;
			}
			else
			{
				throw new IllegalArgumentException(chestName);
			}
		}
	}
	
	public boolean hasChestName(final String chestName)
	{
		synchronized (_mapsMonitor)
		{
			return _virtualInventoriesMap.containsKey(chestName.toLowerCase());
		}
	}

	public String getLanguage()
	{
		return ConfigurationUtils.getConfigValueByType(Configurations.LANGUAGE, String.class);
	}

	public String getConfigList()
	{
		final StringBuilder strBld = new StringBuilder();

		strBld.append(MessagesUtils.getWithColor(Messages.CONFIG_LIST, ChatColor.YELLOW));
		boolean isFirst = true;

		for (final Configurations conf : Configurations.values())
		{
			if (!isFirst)
			{
				strBld.append(ChatColor.YELLOW).append(Constants.DELIMITER).append(Constants.EMPTY_STRING);
			}
			else
			{
				isFirst = false;
			}

			strBld.append(ChatColor.AQUA).append(conf.name).append(ChatColor.YELLOW).append(Constants.EQUAL);
			
			if (conf.type == Boolean.class)
			{
				final boolean value = ConfigurationUtils.getConfigValueByType(conf, Boolean.class);
				if (value)
				{
					strBld.append(ChatColor.GREEN);
				}
				else
				{
					strBld.append(ChatColor.RED);
				}
				strBld.append(value);
			}
			else
			{
				strBld.append(ChatColor.WHITE).append(ConfigurationUtils.getConfigValueByType(conf, String.class));
			}
		}

		return strBld.toString();
	}

	public void displayCompleteHelpMessage(final Player player)
	{
		player.sendMessage(MessagesUtils.getWithColor(Messages.HELP_MAIN, ChatColor.YELLOW));
		player.sendMessage(MessagesUtils.getWithColor(Messages.HELP_ADMINS, ChatColor.YELLOW));
		player.sendMessage(MessagesUtils.getWithColor(Messages.HELP_PLAYERS, ChatColor.YELLOW));
		player.sendMessage(MessagesUtils.getWithColor(Messages.HELP_CHEST, ChatColor.YELLOW));
	}

	public String getAliasesListForCommand(final Commands command)
	{
		final StringBuilder strBld = new StringBuilder();
		strBld.append(MessagesUtils.getWithColors(Messages.ALIASES_AVAILABLE, ChatColor.YELLOW, ChatColor.AQUA,
				command.name()));
		boolean isFirst = true;
		
		for (final String str : command.list)
		{
			if (!isFirst)
			{
				strBld.append(Constants.DELIMITER).append(Constants.EMPTY_STRING);
			}
			else
			{
				isFirst = false;
			}
			strBld.append(ChatColor.AQUA).append(str).append(ChatColor.YELLOW);
		}

		return strBld.toString();
	}
	
	public String getChestDetails(final String name)
	{
		final String chestName = name.toLowerCase();
		final StringBuilder strBld = new StringBuilder();
		
		strBld.append(MessagesUtils.getWithColors(Messages.CHEST_DETAILS, ChatColor.YELLOW, ChatColor.AQUA, chestName));
		
		VirtualInventory vInv = null;
		synchronized (_mapsMonitor)
		{
			vInv = _virtualInventoriesMap.get(chestName);
		}

		if (vInv != null)
		{
			strBld.append(ChatColor.WHITE).append(vInv.getDataString());
		}		
		
		return strBld.toString();
	}

	/* PRIVATE METHODS */

	private boolean loadLinkedChestData()
	{
		if (_pluginData != null)
		{
			if (!_pluginData.isEmpty())
			{
				for (final Object key : _pluginData.keySet())
				{
					final String chestName = String.valueOf(key);
					if (StringUtils.isValidChestName(chestName))
					{
						final String lowerChestName = chestName.toLowerCase();
						final String value = _pluginData.getProperty(lowerChestName);
						if (StringUtils.isNotBlank(value))
						{
							final Location loc = StringUtils.parseLocation(this, value);
							if (loc != null)
							{
								final int indexDelimiter = value.lastIndexOf(Constants.DELIMITER);
								if (indexDelimiter > 0)
								{
									final String chestOptions = value.substring(indexDelimiter + 1);
									final boolean isAdminChest = Commands.ADMIN_CHEST.contains(chestOptions);
									final VirtualInventory virtualInv = new VirtualInventory(loc, isAdminChest);
									setVirtualInventory(lowerChestName, virtualInv);
								}
								else
								{
									_LOGGER.severe("Invalid location format for chestName '" + lowerChestName + "' : "
											+ value);
								}
							}
							else
							{
								_LOGGER.severe("Invalid location format for chestName '" + lowerChestName + "' : "
										+ value);
							}
						}
						else
						{
							_LOGGER.warning("Ignoring blank value for entry : " + chestName);
						}
					}
					else
					{
						_LOGGER.warning("Ignoring blank name entry : " + chestName);
					}
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}
}
